"""
classes for the GML types
$Id: GMLTypes.py,v 1.5 2000/08/29 09:52:58 konrad Exp $
"""

import Stackmachine, Ops, Stack, Exceptions
SM=Stackmachine

class AnyGMLType:
    "should match any type"

class Closure(AnyGMLType):
    def __init__(self, symbols, ilist):
        "takes a symbol table (which won't be modified) and an instr-list"
        self.symbols=symbols.copy()
        self.ilist=ilist
        self.machine=SM.Stackmachine()
        
    def run(self, stack):
        "runs the ilist on the given stack, possibly leaving results there"
        self.machine.setStackAndSymbols(stack, self.symbols.copy())
        #can we avoid copying a hash at each function call?
        self.machine.run(self.ilist)

    def stack_call(self, arglist):
        """
        call the closure with a stack built on arglist (which may be modif.)
        returns the stack top index
        Example:
          clo comes from {/z /y /x   x y addi   y z subi }
          st=[1,2,7]
          clo(st) ==> 2
          #st==[3,-5,7]
        """
        stack=Stack.Stack(arglist)
        self.run(stack)
        return stack.top

    def __call__(self, *args):
        stack=Stack.Stack(list(args))
##        print "Calling Closure %s with %s"%(self,stack.items)
        self.run(stack)
        return tuple(stack.items[:stack.top])
        

class Array(AnyGMLType):
    def __init__(self, symbols, ilist):
        "takes a symbol table (won't be modified) and an ilist"
        machine=SM.Stackmachine(None, symbols.copy())
        machine.run(ilist)
        self.items=machine.stack.popall()
        self.arraylength=len(self.items)
        
    def get(self, key, coor=(0,0)):
        return self.items[key]

    def length(self):
        return self.arraylength

    def asList(self):
        return self.items

    def asTuple(self):
        return tuple(self.items)

class Point(AnyGMLType):
    def __init__(self, x,y,z):
        self.x=x
        self.y=y
        self.z=z
    def __str__(self):
        return "<%f, %f, %f>"%(self.x, self.y, self.z)
    def asList(self):
        return [self.x,self.y,self.z]
    def asTuple(self):
        return (self.x,self.y,self.z)

class Boolean(AnyGMLType):
    def __init__(self, b):
        self.value=b
    def __str__(self):
        if self.value: return "TRUE"
        else: return "FALSE"

