#---------------------------------------------------------------------------
# Cobra Team - 2000 - ICFP'200 Programming Contest
#---------------------------------------------------------------------------

from MathLib import *

dbgRender=0

#---------------------------------------------------------------------------
# General implementation not:
# An intersection is (t, intersectionPoint, normal, (u,v,face), object)
# Intersection functions return a list of such intersections.
#---------------------------------------------------------------------------

#---------------------------------------------------------------------------

def solveEq1(a,b):
    # a x + b = 0
    if a==0: return Infinite # XXX: check overflow
    return -b/a

def solveEq2(a,b,c):
    # XXX: does not work for a==0
    #if a==0.0:
    #	return solveEq1(b,c)
    delta=b*b-4*a*c
    if delta<0: return None
    sqrtDelta=math.sqrt(delta)
    s1=(-b+sqrtDelta)/(2*a)
    s2=(-b-sqrtDelta)/(2*a)
    return s1, s2

def cosSinToAngleOver2Pi(lc,ls):
    """lc= l*cos(angle) ; ls=l*sin(angle)
    return an angle in range[0,1]"""
    angle=math.atan2(ls,lc)
    if angle<0: angle=angle+2.0*math.pi
    return angle/(2.0*math.pi)

#---------------------------------------------------------------------------

def planeIntersection(position, direction,
                      reference, cutOffT, justTest=0):
    faceIndex=0
    isPlane=1
    if dbgRender: print "planeIntersection", position, direction
    py=position[1]
    vy=direction[1]
    if isZero(vy): return [] # XXX: assume we are parallel
    t=-py/vy
    intersection=[position[0]+t*direction[0],0.0, position[2]+t*direction[2]]
    u=intersection[0]
    v=intersection[2]
    
    if t<cutOffT: return [] # XXX: is it sure?
    if justTest: return 1

    #if py>0: normal=[0,1,0]
    #else: normal=[0,-1,0]
    normal=[0,1,0]
    result=[(t,intersection,normal,(u,v,faceIndex),reference)]
    return result

#---------------------------------------------------------------------------

def planarSurfaceIntersection(surfaceFunction, advisedFace,
                              position, direction, reference, cutOffT,
                              justTest=0):
    """Intersection with a planar surface.
    surfaceFunction takes (x,z,face) and returns (u',v',face'),normalDir)
    if inside or None otherwise. normalDir is 1.0 or -1.0.
    advisedFace is only here to avoid making 1 function per cube face
    Returns: a list of intersection (or 1 if justTest _and_
    there is intersection)
    """
    if dbgRender: print "planarSurfaceIntersection:", surfaceFunction,\
       advisedFace, position, direction, reference, cutOffT, justTest
    intersectList=planeIntersection(position, direction, reference, cutOffT)
    if len(intersectList)==0:
        return []
    t,intersection,normal,uvFace,object=intersectList[0]
    u,v,unusedFace=uvFace
    newUVFace=surfaceFunction(u,v,advisedFace)
    if dbgRender:
        print "planarSurface t,intersec,normal,uvFace,newUVFace", \
              t,intersection, normal, uvFace, newUVFace
    if newUVFace==None: return []
    normalDir=newUVFace[1]
    normal=scaleVector(normalDir, normal)
    result=[(t, intersection, normal, newUVFace[0], reference)]
    if dbgRender: print "planarSurface result", result
    return result

# Two basic functions for surfaceIntersection
def discFunction(u,v,face):
    """A sphere of radius 1 in 0,0"""
    distance=u*u+v*v
    if distance>1: return None
    return XXX

#---------------------------------------------------------------------------
# For cube, planarSurfaceIntersection related functions

def squareFunction(u,v,face):
    """A square (0,0) -> (1,1)"""
    if 0<=u<=1 and 0<=v<=1: return (u,v,face), 1
    else: return None

def mirroredSquareFunction(u,v,face):
    """A square (0,0) -> (1,1)"""
    if 0<=u<=1 and 0<=v<=1: return (u,v,face), -1
    else: return None

#---------------------------------------------------------------------------
# For disc, planarSurfaceIntersection related Function:

def discCylinderTopFunction(x,z,face):
    """A sphere of radius 1 in 0,0"""
    distance=x*x+z*z
    if distance>1: return None
    return ((x+1)/2.0, (z+1)/2.0, face), +1

def discCylinderBottomFunction(x,z,face):
    """A sphere of radius 1 in 0,0"""
    distance=x*x+z*z
    if distance>1: return None
    return ((x+1)/2.0, (z+1)/2.0, face), -1

# The cone has the same as the disc
discConeBaseFunction=discCylinderTopFunction

# Intersection of a cylinder with just the side part
def cylinderIntersection(position, direction, reference, cutOffT, justTest=0):
    if dbgRender: print "cylinder intersection pos=",position,"dir=",direction
    a=direction[0]*direction[0]+direction[2]*direction[2]
    b=2*direction[0]*position[0]+2*direction[2]*position[2]
    c=position[0]*position[0]+position[2]*position[2]-1

    if dbgRender: print "2nd degree equation:",a,b,c
    if isZero(a): return []
    eqResult=solveEq2(a,b,c)
    if eqResult==None: return []
    t1,t2=eqResult
    point1=ray(position, t1, direction)
    point2=ray(position, t2, direction)
    angle1=cosSinToAngleOver2Pi(point1[2], point1[0])
    angle2=cosSinToAngleOver2Pi(point2[2], point2[0])
    
    normal1=[point1[0], 0, point1[2]]
    normal2=[point2[0], 0, point2[2]]
    if dbgRender:
        print "t1=",t1,"point1=",point1,"angle1=",angle1,"normal1=",normal1
        print "t2=",t2,"point2=",point2,"angle2=",angle2,"normal2=",normal2
    result=[]
    if 0<=point1[1]<=1 and t1>cutOffT:
        result.append((t1, point1, normal1, (angle1, point1[1], 0), reference))
    if 0<=point2[1]<=1 and t2>cutOffT:
        result.append((t2, point2, normal2, (angle2, point2[1], 0), reference))
    if justTest and len(result)>0: return 1
    if dbgRender: print "cylinderIntersection result=",result
    return result

#---------------------------------------------------------------------------

Sqrt2=math.sqrt(2)

def computeConeNormal(x,z):
    y=math.sqrt(x*x+z*z)
    if isZero(y): return [0, -1, 0]
    return normalizeVector([x/y,-1,z/y])

# Intersection with just the side part
def coneIntersection(position, direction, reference, cutOffT, justTest=0):
    if dbgRender: print "code intersection pos=",position,"dir=",direction
    a=direction[0]*direction[0]+direction[2]*direction[2]\
       -direction[1]*direction[1]
    b=2*direction[0]*position[0]+2*direction[2]*position[2]\
       -2*direction[1]*position[1]
    c=position[0]*position[0]+position[2]*position[2]\
       -position[1]*position[1]

    if dbgRender: print "2nd degree equation:",a,b,c    
    if isZero(a): return []
    eqResult=solveEq2(a,b,c)
    if eqResult==None: return []
    t1,t2=eqResult
    point1=ray(position, t1, direction)
    point2=ray(position, t2, direction)
    angle1=cosSinToAngleOver2Pi(point1[2], point1[0])
    angle2=cosSinToAngleOver2Pi(point2[2], point2[0])
    
    normal1=computeConeNormal(point1[0], point1[2])
    normal2=computeConeNormal(point2[0], point2[2])
    if dbgRender:
        print "t1=",t1,"point1=",point1,"angle1=",angle1,"normal1=",normal1
        print "t2=",t2,"point2=",point2,"angle2=",angle2,"normal2=",normal2

    result=[]
    if 0<=point1[1]<=1 and t1>cutOffT:
        result.append((t1, point1, normal1, (angle1, point1[1], 0), reference))
    if 0<=point2[1]<=1 and t2>cutOffT:
        result.append((t2, point2, normal2, (angle2, point2[1], 0), reference))
    if justTest and len(result)>0: return 1
    if dbgRender: print "coneIntersection result=",result    
    return result

#---------------------------------------------------------------------------

def sphereXYZToUV(point):
    v=(point[1]+1.0)/2.0
    r=1-point[1]*point[1]
    if isZero(r) or r<0: return (0.5, v) #XXX: correct bailout?
    r=math.sqrt(r)
    c=point[2]
    s=point[0]
    return cosSinToAngleOver2Pi(c,s),v

# Unused.
#(XXX: Feature of this version: cutOffT can't be !=0 with justTest=1)
def oldSphereIntersection(position, direction, reference, cutOffT, justTest=0):
    if dbgRender: print "sphereIntersection", position, direction
    normDirection=normVector(direction) #XXX: what if Infinite?
    if isZero(normDirection): return []
    tca=-dotProdVector(position, direction)/normDirection
    squareEyeSphere=dotProdVector(position,position)
    if dbgRender: print "t_value ortho, direction:", tca, direction
    if tca<0: return []
    squareRayDistance=squareEyeSphere-tca*tca
    if dbgRender: print "squareRayDistance", squareRayDistance
    if squareRayDistance>1: return []
    if justTest: return 1
    
    squareThc=1-squareRayDistance
    if squareThc<0: return [] #XXX: should not happen
    thc=math.sqrt(squareThc)
    t1=tca-thc
    t2=tca+thc    
    point1=ray(position,t1,direction)
    point2=ray(position,t2,direction)
    if squareEyeSphere>1: normal1=point1[:] ; normal2=point2[:]
    else: normal1=scaleVector(-1,point1); normal2=scaleVector(-1,point2)
    u1,v1=sphereXYZToUV(point1)
    u2,v2=sphereXYZToUV(point2)
    result=[]
    if t1>cutOffT: result.append((t1,point1,normal1,(u1,v1,0), reference))
    if t2>cutOffT: result.append((t2,point2,normal2,(u2,v2,0), reference))
    if dbgRender: print "sphereIntersectResult=", result
    if justTest and len(result)>0: return 1    
    return result

#---------------------------------------------------------------------------
	
def newSphereIntersection(position, direction, reference, cutOffT, justTest=0):
    # Convert to 2nd order equation and solve it.
    if dbgRender: print "newSphereIntersection", position, direction
    vPos=Vector(position)
    vDir=Vector(direction)
    a=(vDir&vDir)
    if isZero(a): return [] 
    b=2*(vPos & vDir)
    c=(vPos & vPos)-1
    eqResult=solveEq2(a,b,c)
    
    if dbgRender: print "t solving:", eqResult
    if eqResult==None: return []
    t1,t2=eqResult
    point1=ray(position, t1, direction)
    point2=ray(position, t2, direction)
    if c<0: normal1=-1*Vector(point1) ; normal2=-1*Vector(point2) #interior
    else: normal1=Vector(point1) ; normal2=Vector(point2) #exterior
    normal1=normal1.data[:]
    normal2=normal2.data[:]
    u1,v1=sphereXYZToUV(point1)
    u2,v2=sphereXYZToUV(point2)
    result=[]
    if t1>cutOffT: result.append((t1,point1,normal1,(u1,v1,0), reference))
    if t2>cutOffT: result.append((t2,point2,normal2,(u2,v2,0), reference))
    if dbgRender: print "sphereIntersectResult=", result
    if justTest and len(result)>0: return 1    
    return result

#---------------------------------------------------------------------------

sphereIntersection=newSphereIntersection

#---------------------------------------------------------------------------

def testBaseIntersection():
    print "angle(1,1) ->", cosSinToAngleOver2Pi(1,1)
    print "angle(-1,1) ->", cosSinToAngleOver2Pi(-1,1)
    print "angle(-1,-1) ->", cosSinToAngleOver2Pi(-1,-1)
    print "angle(1,-1) ->", cosSinToAngleOver2Pi(1,-1)

#---------------------------------------------------------------------------

if __name__=="__main__":
    testBaseIntersection()

#---------------------------------------------------------------------------
