"""
Bytecode, typename etc. Prettyprinters
$Id: PrettyPrint.py,v 1.6 2000/08/29 10:51:48 konrad Exp $
(C) 2000 Konrad Anton
"""

import Stackmachine
import Ops
import GMLTypes
from types import *
SM=Stackmachine


class InstrListPrinter:
    def __init__(self, ilist):
        self.closures_to_show={}
        self.ilist=ilist

    def printInstr( self, instr, nesting=0 ):
        "bytecode pretty printer"
        (line,col)= instr[0]
        opcode=instr[1]
        print ">"*nesting +"%d(col %d): %s "%(line,col, SM.instrname[opcode]) ,
        if opcode in (SM.APPLY,SM.ITE):
            print 
        elif opcode == SM.PUSHC:
            va=instr[2]
            print " %s (:%s)" %( str(va), typename(va))
            if isinstance(va, GMLTypes.Closure):
                self.closures_to_show[str(va)]=(va,nesting)
            elif isinstance(va, GMLTypes.Array):
                print "          in array:", va.items
        elif opcode in ( SM.CLOSURE, SM.ARRAY ) :
            print
            for i in instr[2]:
                self.printInstr( i, nesting+1 )
        else:
            print repr(instr[2])

    def printAll(self):
        print "Stack machine instruction listing:"
        for instr in self.ilist:
            self.printInstr(instr, 0)
            
        print "Instant closures:"
        for closure_level in range(100):
            for clo, lev in self.closures_to_show.values():
                if lev == closure_level:
                    print "* Closure",str(clo), "is:"
                    print "    symbols:",str(clo.symbols)
                    map(lambda i,pri=self.printInstr,lev=lev: \
                            pri(i,lev+1),
                        clo.ilist)


def typename( v ):
    """
    if v is a primitive, return its name,
    if v is an instance of a known class, return the classname,
    if v is a type, return the type's name
    """
    tt=type(v)
    if tt is TypeType :
        tt = v # someone passed a raw type
        
    if tt is IntType or tt is LongType:
        return "Int"
    elif tt is StringType:
        return "String"
    elif tt is FloatType:
        return "Float"
    elif tt is InstanceType:
        if isinstance(v, GMLTypes.Boolean):
            return "Boolean"
        elif isinstance(v, GMLTypes.Point):
            return "Point"
        elif isinstance(v, GMLTypes.Closure):
            return "Closure"
        elif isinstance(v, GMLTypes.Array):
            return "Array"
    elif tt is ClassType:
        if v is GMLTypes.Boolean: return "Boolean"
        elif v is GMLTypes.Point: return "Point"
        elif v is GMLTypes.Closure: return "Closure"
        elif v is GMLTypes.Array: return "Closure"
    else:
        return repr(type(v))
    return "THIS TYPE IS ILLEGAL. GIVE IT BACK."
