#! /usr/bin/env python
#---------------------------------------------------------------------------
# Cobra Team - 2000 - ICFP'200 Programming Contest
#
# Cedric Adjih - adjih@crepuscule.com
#---------------------------------------------------------------------------
#
# A "Simple" Renderer
#
# Design:
# - The renderer throw a simple ray and compute all the intersections
#   with all the objects of the scene. Acne is avoided by passing in
#   argument a pointer to one element to avoid.
# - No optimisations
#
# Structure: it has grow wildly so it probably happens to be the worse
# Python code I wrote :-)
# - The factory is used to create a scene
# - At rendering time, the transform are propagated through the whole scene
#   an a "rendering scene" is returned.
# - It was expected to make a C++ backend so the rendering scene is just
#   a list with tags like "scene", "union", "transform". Using classes
#   would be signicantly cleaner, but no time for cleaning it.
#
#---------------------------------------------------------------------------

# See below for import from Parser

#---------------------------------------------------------------------------

from MathLib import *
from Intersection import *

import Intersection
import string

#---------------------------------------------------------------------------

#ZZZ! remove this
#import traceback, types, StringIO, sys

def getTraceback():
    stderr=sys.stderr
    errMessage=StringIO.StringIO()
    sys.stderr=errMessage
    traceback.print_exc()
    sys.stderr=stderr
    return errMessage.getvalue()

#---------------------------------------------------------------------------

class AbstractSceneFactory:
    def isSceneObject(self, object):                       undefined
    def isLightObject(self, object):                       undefined
    
    def makeSphere(self, surfaceFunction):                 undefined
    def makeCube(self, surfaceFunction):                   undefined 
    def makeCylinder(self, surfaceFunction):               undefined 
    def makeCone(self, surfaceFunction):                   undefined 
    def makePlane(self, surfaceFunction):                  undefined
    
    def applyTranslation(self, sceneObject, rx, ry, rz):   undefined
    def applyScale(self, sceneObject, rx, ry, rz):         undefined
    def applyUScale(self, sceneObject, rx, ry, rz):        undefined
    def applyRotateX(self, sceneObject, angle):            undefined
    def applyRotateY(self, sceneObject, angle):            undefined
    def applyRotateZ(self, sceneObject, angle):            undefined

    def makeLight(self, direction, color):                 undefined
    def makePointLight(self, direction, color):            undefined
    def makeSpotLight(self, pos, dir, color, cutoff, exp): undefined
    
    def makeUnion(self, object1, object2):                 undefined
    def makeIntersect(self, object1, object2):             undefined
    def makeDifference(self, object1, object2):            undefined

#---------------------------------------------------------------------------

GeomPrimitive=["sphere","cube","cylinder","cone","plane", # GML Primitive
               "planarSurface" # Internal primitive
               ]

class SimpleSceneObject:
    def __init__(self,*data):
        self.data=data
    def __getitem__(self, index):
        return self.data[index]
    def __len__(self):
        return len(self.data)
    def getSceneToRender(self):
        currentTransform=Transform()
        return self.getSceneWithTransform(currentTransform)
    def getSceneWithTransform(self, transform):
        """Transform everything as a list, but applying Transform
        recursively so that each object has its associated matrix"""
        if self[0]=="transform":
	    newTransform=transform*self[1]
            return self[2].getSceneWithTransform(newTransform)
        elif self[0] in ["union", "intersect", "difference"]:
            return [self[0], self[1].getSceneWithTransform(transform),
                    self[2].getSceneWithTransform(transform)]
        else:
            assert self[0] in GeomPrimitive
            return [self[0], self[1], transform]+list(self.data[2:])
    def __repr__(self):
	return "<SimpleScene"+repr(self.data)+">"

class SimpleLightObject:
    def __init__(self, *data):
        self.data=data
    def __getitem__(self, index):
        return self.data[index]

class SimpleSceneFactory:
    """Note: each sceneObject must have checks first"""
    def isSceneObject(self, object):
        return isinstance(object, SimpleSceneObject)
    def isLight(self, object):
        return isinstance(object, SimpleLightObject)

    #--- Geometric primitives

    def makeSphere(self, surfaceFunction):
        return SimpleSceneObject("sphere", surfaceFunction)

    def _makeSquare(self, surfaceFunction, faceIndex, mirrored):
        if mirrored: uvFunction=mirroredSquareFunction
        else: uvFunction=squareFunction
        return SimpleSceneObject("planarSurface", surfaceFunction,
                                 uvFunction, faceIndex)
    
    def makeCube(self, surfaceFunction):
        f=surfaceFunction
        face5=self._makeSquare(f,5,1)
        face0=self.applyRotateX(self._makeSquare(f,0,0), -90)
        t=self.applyRotateX(self._makeSquare(f,1,1), -90)        
        face1=self.applyTranslate(t, 0,0,1)
        t=self.applyRotateX(self._makeSquare(f,2,1), -90)
        face2=self.applyRotateY(t,-90)
        t=self.applyRotateX(self._makeSquare(f,3,0), -90)
        t=self.applyRotateY(t, -90)
        face3=self.applyTranslate(t,1,0,0)
        face4=self.applyTranslate(self._makeSquare(f,4,0), 0,1,0)
        return reduce(self.makeUnion, [face0,face1,face2,face3,face4,face5])
        #return face5

        return SimpleSceneObject("cube", surfaceFunction)
    def makeCylinder(self, surfaceFunction): 
        side=SimpleSceneObject("cylinder", surfaceFunction)
        disc=SimpleSceneObject("planarSurface", surfaceFunction,
                               discCylinderTopFunction, 1)
        top=self.applyTranslate(disc, 0,1,0)
        bottom=SimpleSceneObject("planarSurface", surfaceFunction,
                                 discCylinderBottomFunction, 2)
        #return self.makeUnion(top,bottom)
        return self.makeUnion(side, self.makeUnion(top,bottom))
    def makeCone(self, surfaceFunction):
        side=SimpleSceneObject("cone", surfaceFunction)
        t=SimpleSceneObject("planarSurface", surfaceFunction,
                            discConeBaseFunction, 1)
        base=self.applyTranslate(t, 0,1,0)
        return self.makeUnion(side, base)
    def makePlane(self, surfaceFunction):
        return SimpleSceneObject("plane", surfaceFunction)

    #--- Transforms
    
    def applyTranslation(self, sceneObject, rx, ry, rz):
        transform=Transform()
        transform.toTranslation([rx, ry, rz])
        return SimpleSceneObject("transform", transform, sceneObject)
    
    applyTranslate=applyTranslation
    
    def applyScale(self, sceneObject, rx, ry, rz):
        transform=Transform()
        transform.toScale([rx, ry, rz])
        return SimpleSceneObject("transform", transform, sceneObject)
    
    def applyUScale(self, sceneObject, scale):
        transform=Transform()
        transform.toScale([scale, scale, scale])
        return SimpleSceneObject("transform", transform, sceneObject)
        
    def applyRotateX(self, sceneObject, angle):
        transform=Transform()
        transform.toRotationX(angle)
        return SimpleSceneObject("transform", transform, sceneObject)
    def applyRotateY(self, sceneObject, angle):
        transform=Transform()
        transform.toRotationY(angle)
        return SimpleSceneObject("transform", transform, sceneObject)        
    def applyRotateZ(self, sceneObject, angle):
        transform=Transform()
        transform.toRotationZ(angle)
        return SimpleSceneObject("transform", transform, sceneObject)        

    #--- Lights
    
    def makeLight(self, direction, color):
        return SimpleLightObject("light", direction, color)
    def makePointLight(self, position, color):
        return SimpleLightObject("pointLight", position, color)
    def makeSpotLight(self, pos, dir, color, cutoff, exp):
        return SimpleLightObject("spotLight", pos, dir, color, cutoff, exp)
    
    def makeUnion(self, object1, object2):
        return SimpleSceneObject("union", object1, object2)
    def makeIntersect(self, object1, object2):
        return SimpleSceneObject("intersect", object1, object2)
    def makeDifference(self, object1, object2):
        return SimpleSceneObject("difference", object1, object2)

#---------------------------------------------------------------------------

dbgBadNormal=0

IdxT=0
IdxIntersection=1
IdxNormal=2
IdxUV=3
IdxObject=4

ObjType=0
ObjFunction=1
ObjTransform=2
ObjUVFunction=3 # Only planarSurface
ObjFaceIndex=4  # Only planarSurface

LightType=0
LightDirection=1
LightColor=2
LightPositionPointLight=1

LightPositionSpotLight=1


def pointToRGB(v):
    result=""
    for c in v:
	c=int(c*255)
	if c<0: c=0
	if c>255: c=255
	result=result+chr(c)
    return result

# dbgData: (position, direction, maxDepth, cutOffT, avoid_surface)

# c -> Cast one ray <python tuple x,y>
# s -> SubSample <coefficient>
# r -> Reference <Image>
# i -> Ignore any runtime error
getopt_args="c:s:r:iu"

optionList=[]
##<konrad>
def setOptionList(olist):
    global optionList
    optionList=olist # can be global since it will not change at runtime ;)
##</konrad>

#---------------------------------------------------------------------------

def computeAttenuation(lightSource, vQ):
    unusedType,pos, dest, color, cutoff, exp=lightSource
    if 0:
        print "pos=", pos
        print "dir=", dir
        print "color=", color
        print "cutoff=", cutoff
        print "exp=", exp
        print lightSource.data, vQ
    vPos=Vector(pos)
    vDest=Vector(dest)
    v1=vDest-vPos
    v2=vQ-vPos
    dotProd=(v1&v2)
    normProd=(normVector(v1.data)*normVector(v2.data))
    if dotProd<0 or isZero(dotProd) or isZero(normProd): return 0.0
    cosValue=dotProd/normProd
    minCosAngle=math.cos(degreeToRadian(cutoff))
    #print degreeToRadian(cutoff), cosValue, minCosAngle
    if cosValue<minCosAngle: return 0.0
    #print math.log(cosValue), exp
    #return 0.0
    attenuation=math.exp(math.log(cosValue)*exp)
    if 0: print "attenuation",attenuation    
    return attenuation
    exitNow

#---------------------------------------------------------------------------
#

def differenceBoolFunction(a1, a2):
    return a1 and not a2

def intersectionBoolFunction(a1, a2):
    return a1 and a2

def sortAndMerge(u1, u2, f):
    #ZZZ: print "sortAndMerge"
    u1.sort()
    u2.sort()
    state=0
    stateU1=0
    stateU2=0
    result=[]
    while 1:
        if len(u1)>0 and len(u2)>0:
            if u1[0][0]<u2[0][0]: choice=1
            else: choice=2
        if len(u1)==0: choice=2
        if len(u2)==0: choice=1
        if len(u1)==0 and len(u2)==0:
            return result
        if choice==1:
            stateU1=stateU1^1
            previousState=state
            state=f(stateU1, stateU2)
            if previousState!=state:
                result.append(u1[0])
            u1=u1[1:]
        else:
            stateU2=stateU2^1
            previousState=state
            state=f(stateU1, stateU2)
            if previousState!=state:
                result.append(u2[0])
            u2=u2[1:]
    

def computeDifference(u1, u2):
    return sortAndMerge(u1,u2,differenceBoolFunction)

def computeIntersection(u1, u2):
    return sortAndMerge(u1,u2,intersectionBoolFunction)

def testSortAndMerge():
    u=[[1],                     [4], [5],      [7]]
    v=[     [2], [3], [3.5],             [6]]
    print computeDifference(u,v)
    print computeIntersection(u,v)
    print computeDifference(v,u)

#testSortAndMerge()
               
#---------------------------------------------------------------------------

class SimpleRenderer:

    #--------------------------------------------------
    
    def __init__(self, ambientColor, lightList, scene):

        self.ambientColor=ambientColor
        self.scene=scene.getSceneToRender()
        self.lightList=lightList
        self.purePython=0
        self.exceptionCounter=0
        self.exceptionRaytraceCounter=0
        self.safe=0
        self.tooSafe=0 # to be true: don't catch GML exceptions
        self.subSample=1
        self.referenceImage=None
        self.setPseudoOptionList(optionList)
        self.oneShot=None
        self.lastGMLException=None

    #--------------------------------------------------

    def getRenderingScene(self): return self.scene

    #--------------------------------------------------

    def setPurePython(self, purePython):
        self.purePython=purePython

    #--------------------------------------------------

    def setPseudoOptionList(self, data):
        i=0
        result=[]
        while i<len(data):
            if data[i] in ["-s","-r"]:
                result.append((data[i], data[i+1]))
                i=i+2
            else:
                result.append((data[i],))
                i=i+1
        self.setOptionList(result)

    #--------------------------------------------------

    def setOptionList(self, data):
        for i in data:
            if i[0]=="-s": self.subSample=string.atoi(i[1])
            elif i[0]=="-r": self.referenceImage=i[1]
            elif i[0]=="-i": self.safe=1
            elif i[0]=="-u": self.tooSafe=1
            elif i[0]=="-c": self.oneShot=eval(i[1])

    #--------------------------------------------------
        
    #XXX: check handling of maxDepth
    def prepareRender(self, maxDepth, fov, xSize, ySize, fileName):
        import GMLTypes
        self.maxDepth=maxDepth
        self.fov=fov
        self.xSize=xSize
        self.ySize=ySize
        self.maxX=math.tan(degreeToRadian(fov)/2.0)
        self.maxY=self.maxX*ySize/float(xSize)
        self.pixelSize=2*self.maxX/self.xSize
        self.refX=-self.maxX
        self.refY=self.maxY
        self.fileName=fileName

    #--------------------------------------------------

    def renderAllScene(self):
        if self.referenceImage!=None:
            import Image # You need PIL (Python Imaging Library for this)
            referenceImage=Image.open(self.referenceImage)
        else: referenceImage=None
        f=open(self.fileName,"w")
        f.write("P6\n# Cobra Team\n%d %d\n255\n" \
                % (self.xSize/self.subSample, self.ySize/self.subSample))
        previousColor=chr(0)+chr(0)+chr(0)
        
        for y in range(0, self.ySize, self.subSample):
            for x in range(0, self.xSize, self.subSample):
                if self.oneShot!=None:
                    setDbgRender("ray")
                    x,y=self.oneShot

                color=self.safeGetScreenPixel(x,y,previousColor)
                if not self.tooSafe and self.lastGMLException!=None:
                    print "*** GMLException caught, you can run with '-u' to skip it ***:"
                    e=self.lastGMLException
                    e[0](e[1],e[2],e[3]) # reproduce the exception
                    raise "terminated"

                # Write it
                f.write(color)

                # Sanity check mode
                if referenceImage!=None:
                    #print "yeah"
                    refPixel=referenceImage.getpixel((x,y))
                    ourPixel=tuple(map(ord, color))
                    if refPixel!=ourPixel:
                        print "Different pixels in x,y=",x,y,"ref/our:",\
                              refPixel,ourPixel
		if self.oneShot: return
        f.close()
        print "Finished rendering scene (forceSafe=%d, forceTooSafe(GML)=%d\n"\
              "   GMLException=%d, raytraceException=%d)" \
              % (self.safe, self.tooSafe, self.exceptionCounter,\
              self.exceptionRaytraceCounter)

    #--------------------------------------------------

    def render(self,maxDepth, fov, xSize, ySize, fileName):
        self.prepareRender(maxDepth, fov, xSize, ySize, fileName)
        self.renderAllScene()

    #--------------------------------------------------

    def renderOnePoint(self,maxDepth, fov, xSize, ySize, fileName, x,y):
        self.prepareRender(maxDepth, fov, xSize, ySize, fileName)
        print "color=", self.getScreenPixel(x,y)

    #--------------------------------------------------

    def safeGetScreenPixel(self, x, y, default):
        if not (self.safe or self.tooSafe): return self.getScreenPixel(x,y)
        else:
            try: return self.getScreenPixel(x,y)
            except:
                #print getTraceback() #ZZZZZ
                #exitNow
                self.exceptionRaytraceCounter=self.exceptionRaytraceCounter+1
                return default

    #--------------------------------------------------

    def getScreenPixel(self, x, y):
        xx=self.refX+(x+0.5)*self.pixelSize
        yy=self.refY-(y+0.5)*self.pixelSize
        eyePos=[0.0,0.0,-1.0]
        eyeDirection=[xx,yy,1]
        if dbgRender: print eyePos, eyeDirection
        color=self.raytrace(eyePos, eyeDirection, self.maxDepth, 0.0, None)
        return pointToRGB(color)

    #--------------------------------------------------

    def safeCallFunction(self, function, a,b,c):
        if not (self.safe or self.tooSafe):
            color,kd,ks,phong=function(a,b,c)
        else:
            try:
                color,kd,ks,phong=function(a,b,c)
            except:
                self.exceptionCounter=self.exceptionCounter+1
                self.lastGMLException=(function,a,b,c)
                return [0.0,0.0,0.0], 0.0, 0.0, 0.0
        if not self.purePython: color=color.asTuple()
        return color,kd,ks,phong

    #--------------------------------------------------

    def raytrace(self, position, direction, remainingDepth, cutOffT,
		 ignoreSurface):
        #CRASH
	global dbgRender, dbgData, dbgObjectType
        if dbgRender: print "--- raytrace: pos, dir:\n",position,direction,remainingDepth
        intersectionList=self.findAllIntersection(self.scene,
                  position, direction, cutOffT, ignoreSurface)
        if dbgRender: print "--- raytrace: pos, dir, depthLeft,LF,intersec:\n",position,direction,remainingDepth,"\n",intersectionList
        intersectionList.sort()
        if len(intersectionList)==0:
	    if dbgRender: print "--- end raytrace, no intersec, color=(0,0,0)"
	    return [0.0, 0.0, 0.0]
        if dbgRender: print "--color calculation:"
        closest=intersectionList[0]
        sceneObject=closest[IdxObject]
        point=closest[IdxIntersection]
        u,v,face=closest[IdxUV]
        function=sceneObject[ObjFunction]
        if not self.purePython:
            color,kd,ks,phong=self.safeCallFunction(function,
                                     int(face),float(u),float(v))
        else:
            color,kd,ks,phong=self.safeCallFunction(function, u,v,face)

        vPoint=Vector(point)
        direction=Vector(direction).normalize()
        normal=Vector(closest[IdxNormal]).normalize()
        negDir=-1*direction
        vectTh=negDir - (negDir&normal)*normal
        vectS=negDir - 2*vectTh
	vectS=vectS.normalize() #XXX: should already be normalized, anyway
        
        pointColor=kd*Vector(self.ambientColor)
        phongColor=Vector([0,0,0])
        baseColor=Vector([0,0,0])

	if dbgRender=="ray":
	    print "closest", closest
	    print "color,kd,ks,phong=func(u=%s,v=%s):" % (u,v),\
		  color,kd,ks,phong
	    print "dir=",direction,"normal=",normal,"vectTh=",vectTh,\
		  "vectS=",vectS
	    print "pointColor,phongColor,baseColor",pointColor,phongColor,\
		  baseColor

        #XXX: check no >1.0 color overflow
        for lightSource in self.lightList:

            lightSourceLightColor=Vector(lightSource[LightColor])

            # Compute light direction and attenuation
            if lightSource[LightType]=="light":
                lightDir=lightSource[LightDirection]
                attenuation=1.0
                maxT=OmegaSquare
                
            elif lightSource[LightType]=="pointLight":
                lightDir=(vPoint-list(lightSource[LightPositionPointLight]))
                squareLightDistance=lightDir&lightDir
                attenuation=100.0/(99.0+squareLightDistance)
                #maxT=1.0
                if dbgRender: print "light:", lightSource.data, "squareDist=",squareLightDistance, "attenuation=", attenuation
                lightDir=lightDir.data[:] # avoid breaking other code...
                maxT=normVector(lightDir)
                
            elif lightSource[LightType]=="spotLight":

                lightDir=(vPoint-list(lightSource[LightPositionPointLight]))
                squareLightDistance=lightDir&lightDir
                attenuation=100.0/(99.0+squareLightDistance)
                #maxT=1.0
                attenuationAngle=computeAttenuation(lightSource, vPoint)
                if dbgRender: print "light:", lightSource.data, "squareDist=",squareLightDistance, "attenuation=", attenuation
                if dbgRender: print "attenuationAngle", attenuationAngle
                attenuation=attenuation*attenuationAngle

                lightDir=lightDir.data[:] # avoid breaking other code...
                maxT=normVector(lightDir)
                lightSourceLightColor=Vector(lightSource[3])
                
            else: raise "unexpected lightSource type", lightSource

            # Compute phong contribution
            lightColor=attenuation*lightSourceLightColor
            lightDir=-1*Vector(lightDir).safeNormalize(normal) #XXX:normal
            dotProd=(lightDir&normal)
            if dotProd<0: #XXX: check this check
                if dbgRender=="ray":
                    print "light opposite dir", lightSource
                continue
            #XXX: has intersection can be called with [0,0,0]
            if self.hasIntersectionInRange(self.scene, point, lightDir.data,
                                           sceneObject, 0.0, maxT):
                if dbgRender=="ray":
                    print "light shadowed", lightSource
                continue
            vectH=0.5*(negDir+lightDir)
            vectH=vectH.normalize() # XXX: what if norm=0 ?
            dotPhong=vectH&normal
            if dotPhong>=0:
                #XXX: check dot prod <=1.0
                c=math.exp(math.log(dotPhong)*phong) #XXX: better power
                phongColor=phongColor+ks*c*lightColor
            else: c=0
            if dbgRender=="ray":
                print "light: dirToLight, color:", lightDir, lightColor
                print "dotProd, vectH, dotPhong", dotProd, vectH, dotPhong
                print "color increment phong, dotProd^phongExp:", (ks*c)*lightColor, c
                print "color increment base", (kd*dotProd)*lightColor
                print "dotProdCheck H.L H.dir",vectH&negDir,vectH&lightDir
            baseColor=baseColor+(kd*dotProd)*lightColor
        if remainingDepth>0:
            #print vectS, point
            reflColor=ks*Vector(self.raytrace(point, vectS.data, remainingDepth-1,0, sceneObject))
        else: reflColor=Vector([0.0,0.0,0.0])
	
        color=(pointColor+reflColor+phongColor+baseColor)|Vector(color)
            
        
        #XXX!!!! remove this
        if dbgBadNormal and dotProdVector(direction, normal)>0:
            print "-"*30
            print intersectionList
            print direction, normal, dotProdVector(direction, normal)
            #raise "Something is rotten"
            if dbgBadNormal: return [0.0,1.0,0.0]

	if dbgRender=="ray": print "--- end raytrace result=", color.data
        return color.data

    #--------------------------------------------------

    def findAllIntersection(self, scene, position, direction, cutT,
			    ignoreSurface):
        if scene[ObjType] in GeomPrimitive:
            transform=scene[2]
            newPos=transform.inverseMul(position)
            newDir=transform.inverseMul(direction, 0.0)
	    if scene==ignoreSurface: #XXX: SceneId
		return []
            if scene[0]=="sphere":
                tmpResult=sphereIntersection(newPos, newDir, scene,cutT)
            elif scene[0]=="plane":
                tmpResult=planeIntersection(newPos, newDir, scene, cutT)
            elif scene[0]=="cylinder":
                tmpResult=cylinderIntersection(newPos, newDir, scene, cutT)
            elif scene[0]=="cone":
                tmpResult=coneIntersection(newPos, newDir, scene, cutT)
            elif scene[0]=="planarSurface":
                uvCheckFunction=scene[ObjUVFunction]
                #print scene                
                faceIndex=scene[ObjFaceIndex]
                tmpResult=planarSurfaceIntersection(uvCheckFunction, faceIndex,
                                                    newPos,newDir,scene,cutT)
            else: raise "XXX: not implemented", scene
            result=[]
            for inter in tmpResult:
                point=transform.mul(inter[IdxIntersection])
                normal=transform.mul(inter[IdxNormal],0.0)
		if dbgRender=="ray":
		    print "back transform pt",inter[IdxIntersection],"->",point
		    print "back transform normal",inter[IdxNormal],"->",normal
                result.append((inter[0],point,normal,inter[3],inter[4]))
            return result
        else:
            assert scene[0] in ["union", "difference", "intersect"]
            if scene[0]=="union":
                u1=self.findAllIntersection(scene[1],position, direction,
                                            cutT, ignoreSurface)
                u2=self.findAllIntersection(scene[2],position, direction,
                                            cutT, ignoreSurface)
                return u1+u2
            else:
                u1=self.findAllIntersection(scene[1],position, direction,
                                            cutT, ignoreSurface)
                u2=self.findAllIntersection(scene[2],position, direction,
                                            cutT, ignoreSurface)
                if scene[0]=="difference":
                    return computeDifference(u1, u2)
                elif scene[0]=="intersect":
                    return computeIntersection(u1, u2)
                else: raise "XXX: unimplemented"

    #--------------------------------------------------

    def hasIntersectionInRange(self, scene, position, direction,
                               ignoreSurface, minT, maxT):
        if dbgRender=="ray":
            print "hasIntersection in range, minT,maxT=", minT,maxT,\
                  "position=",position,"direction=",direction
        intersectionList=self.findAllIntersection(scene, position,
                                  direction, minT, ignoreSurface)
        if dbgRender:
            print "hasIntersection, intersectionList=", intersectionList
        for intersec in intersectionList:
            if minT<=intersec[IdxT]<=maxT:
                if dbgRender: print "intersection found", intersec
                return 1
        return []

    #--------------------------------------------------
    # Removed, RCS 1.9 file is the last to have it
    #def hasIntersection(self, scene, position, direction, ignoreSurface):

    #--------------------------------------------------

#---------------------------------------------------------------------------

def render(ambientColor, lightList, scene, depth,
           fov, xSize, ySize, outputFile):
    renderer=SimpleRenderer(ambientColor, lightList, scene)
    ##<konrad>
    renderer.setOptionList(optionList) #give it the global options
    ##</konrad>
    renderer.render(depth, fov, xSize, ySize, outputFile)

def pythonicRender(ambientColor, lightList, scene, depth,
                   fov, xSize, ySize, outputFile, oneShot=None):
    #---------------------
    # From Konrad's Parser
    from Parser import GML_Parser, GML_Lexer
    import PrettyPrint
    import Stackmachine, Exceptions
    import getopt
    import Optimizer
    #---------------------
    
    renderer=SimpleRenderer(ambientColor, lightList, scene)
    renderer.setPurePython(1)
    renderer.oneShot=oneShot
    renderer.render(depth, fov, xSize, ySize, outputFile)

#---------------------------------------------------------------------------


def setDbgRender(value, data=None, objectType=None):
    global dbgRender, dbgData, dbgObjectType
    dbgRender=value
    dbgData=data
    dbgObjectType=objectType
    Intersection.dbgRender=value
    Intersection.dbgData=data
    Intersection.dbgObjectType=objectType

#---------------------------------------------------------------------------

def testIntersectionPosDir(scene, listPosDir):
    renderer=SimpleRenderer([1, 1, 1], [], scene)
    renderer.setPurePython(1)
    renderingScene=renderer.getRenderingScene()
    
    for pos,dir in listPosDir:
        print "-"*50
        res=renderer.findAllIntersection(renderingScene, pos, dir, 0, None)
        #res2=renderer.hasIntersection(renderingScene, pos, dir, None)
        res2="-"
        print "findAllIntersection(pos=%s,dir=%s) [hasIntersect=%s]:" \
              % (pos,dir,res2)
        print res

def testIntersection():
    factory=SimpleSceneFactory()
    
    setDbgRender(1)

    # Debug cone
    print "---- Cone"
    cylinder=factory.makeCone(None)
    posDirList=[ ([0,0.5,-2], [0,0,1]) ]
    posDirList=[ ([0,0,-2], [0,0.5,1]) ] 
    #posDirList=[ ([0,0,3], [0,0,-1]) ]
    testIntersectionPosDir(cylinder, posDirList)

    # Debug cylinder
    print "---- Cylinder"
    cylinder=factory.makeCylinder(None)
    posDirList=[]
    for x,z in [(1,1), (-1,1), (-1,-1), (1,-1)]:
        posDirList.append( ([2*x,0.3,2*z], [-x,0,-z]) )
    posDirList=[ ([0,0,-2], [0,0.5,1]) ]
    posDirList=[ ([0,0.3,-2], [0,0,1]) ] 
    posDirList=[ ([0,0,3], [0,0,-1]) ]
    testIntersectionPosDir(cylinder, posDirList)
    
    # Debug cube:
    print "---- Cube"
    cube=factory.makeCube(None)
    posDirList=[]
    posDirList=[
        ([0.25,0.3,-2], [0,0,1]), # for face 0 and face 1
        ([3,0.3,0.25], [-1,0,0]), # for face 2 and face 3
        ([0.25,2,0.3], [0,-1,0]) # for face 4 and face 5
        ]
    testIntersectionPosDir(cube, posDirList)

if __name__=="__main__":
    testIntersection()
    
#---------------------------------------------------------------------------

